defmodule TmunitEx.GbxRemote do
  use Connection

  def child_spec(args) do
    %{
      id: __MODULE__,
      start: {__MODULE__, :start_link, [args]},
      type: :supervisor
    }
  end

  # Client

  def start_link([host, port, timeout, opts]) do
    Connection.start_link(__MODULE__, {host, port, opts, timeout})
  end

  def start([host, port, opts, timeout]) do
    Connection.start(__MODULE__, {host, port, opts, timeout})
  end

  def send(conn, %XMLRPC.MethodCall{method_name: _name, params: params} = req, timeout \\ 5000) when is_list(params) do
    {:ok, xml} = req
      |> XMLRPC.encode
    Connection.call(conn, {:send, xml}, timeout)
  end

  def recv(conn, timeout \\ 5000) do
    Connection.call(conn, :recv, timeout)
  end

  # Callbacks

  @impl true
  def init({host, port, timeout, opts}) do
    state = %{host: host, port: port, opts: opts, timeout: timeout, sock: nil}
    {:connect, :init, state}
  end

  @impl true
  def connect(_, %{sock: nil, host: host, port: port, timeout: timeout, opts: opts} = state) do
    case :gen_tcp.connect(host, port, [active: false] ++ opts, timeout) do
      {:ok, sock} ->
        {:ok, len} = :gen_tcp.recv(sock, 4, timeout)
        len = :binary.decode_unsigned(len, :little)
        if len > 64 do
          {:backoff, 1000, state}
        end

        {:ok, protocol} = :gen_tcp.recv(sock, len, timeout)
        if protocol != "GBXRemote 2" do
          {:backoff, 1000, state}
        end

        {:ok, %{state | sock: sock}}

      {:error, _reason} ->
        {:backoff, 1000, state}
    end
  end

  @impl true
  def handle_call({:send, xml}, _from, %{sock: sock, reqhandle: reqhandle} = state) do
    reqhandle = reqhandle + 1
    breqhandle = <<reqhandle::little-integer-size(32)>>
    len = <<String.length(xml)::little-integer-size(32)>>

    query = len <> breqhandle <> xml

    case :gen_tcp.send(sock, query) do
      :ok ->
        {:reply, :ok, %{state | reqhandle: reqhandle} }
      {:error, _reason} = error ->
        {:disconnect, error, error, state}
    end
  end

  @impl true
  def handle_call(:recv, _from, %{sock: sock, reqhandle: reqhandle, timeout: timeout} = state) do
    case :gen_tcp.recv(sock, 8, timeout) do
      {:ok, data} ->
        {len, recvhandle} = {String.slice(data, 0..3), String.slice(data, 4..7)}
        len = :binary.decode_unsigned(len, :little)

      {:error, _reason} = error ->
        {:error, error, error, state}
    end
  end

  @impl true
  def disconnect(_info, %{sock: sock} = socket) do
    :ok = :gen_tcp.close(sock)
    {:connect, :reconnect, %{socket | sock: nil}}
  end

  @impl true
  def terminate(_reason, %{sock: sock} = socket) do
    :ok = :gen_tcp.close(sock)
    {:ok, %{socket | sock: nil}}
  end
end
